#
# Copyright (c) 2008 Grigori Goronzy <greg@geekmind.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import os, sys, commands
import gobject
import math
import signal
import struct
import socket
import dbus.service
import models
import ioport
import time
import glob
import ichsmbus
import gconf

# Version string
VERSION = "0.8.4"
# Keycodes for media keys
BRIGHTNESS_UP = 225
BRIGHTNESS_DOWN = 224
VOLUME_MUTE = 113
VOLUME_DOWN = 114
VOLUME_UP = 115
# Location of the acpid socket -- is there anyone using a different?
ACPID_SOCKET = "/var/run/acpid.socket"
# Settings for optimized fan control.
# This is slower than the hardware default, but also pretty much on the safe side.
FAN_OPTIMIZED = False   # Start fan control at daemon startup
FAN_CRITICAL = 52.0
FAN_MAX_SPEED = 55.0
FAN_TIMEOUT = 15000
FAN_INTERVAL = 7.0

# Magic numbers for EC/PLL access
PLL_ADDR = 0x69
EC_BASE = 0x380
EC_REG_TEMP = 0xf451
EC_REG_FANPWM = 0xf463
EC_REG_FANSPEED = 0xf466
EC_REG_FEATURE = 0xf4d3
EC_VOLTAGE = 0x66
#--------------------------------------------------------------------------

"""normalize two lists to the length of the longer one"""
def normalize(a, b):
    if len(a) == len(b): return
    if len(b) < len(a):
        a, b = b, a
    delta = len(b)-len(a)
    step = len(a) / delta
    pos = 0
    for i in xrange(0, delta):
        obj = 0
        if i == 0: obj = a[0]
        else: obj = a[i-1]
        a.insert(i, obj)
        pos += step

"""Read an embedded controller (EC) reg"""
def ec_read(adr):
    ioport.outb(adr >> 8, EC_BASE+1)
    ioport.outb(adr & 0xff, EC_BASE+2)
    return ioport.inb(EC_BASE+3)

"""Write an embedded controller (EC) reg"""
def ec_write(adr, data):
    ioport.outb(adr >> 8, EC_BASE+1)
    ioport.outb(adr & 0xff, EC_BASE+2)
    ioport.outb(data, EC_BASE+3)

def ec_pin2port(pin):
    return 0xfc20 + ((pin >> 3) & 0x1f)

def ec_pin2mask(pin):
    return 1 << (pin & 0x07)

def ec_gpio_get(pin):
    port = ec_pin2port(pin)
    mask = ec_pin2mask(pin)
    stat = ec_read(port) & mask
    return bool(stat)

def ec_gpio_set(pin, value):
    port = ec_pin2port(pin)
    mask = ec_pin2mask(pin)
    if bool(value) == True:
        ec_write(port, ec_read(port) | mask)
    else:
        ec_write(port, ec_read(port) & ~mask)

#--------------------------------------------------------------------------

"""Receive and parse ACPI events"""
class AcpiParser:
    def __init__(self, model):
        self.model = model
        # Try to connect, and if it fails, simulate a hangup
        try: self.connect()
        except socket.error: self.hangup_acpi(None, None)
         
    def connect(self):
        # Open connection to acpid socket
        asg = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        asg.connect(ACPID_SOCKET)
        gobject.io_add_watch(asg, gobject.IO_IN, self.read_acpi)
        gobject.io_add_watch(asg, gobject.IO_HUP, self.hangup_acpi)
        log("connected to acpid")
            
    def read_acpi(self, source, cond):
        line = source.recv(1024)
        # Sometimes after the connection gets dropped we still are called...
        if not line: return False
        args = line.strip().split(" ")
        if len(args) != 4: return True
        type, source, code, serial = args
        if type == "hotkey" and source == "ATKD":
            key = int(code, 16)
            log("received hotkey 0x%x" %key)
            self.model.do_action(key)
        return True
    
    """Poll for acpid re-appearance every 10 seconds"""
    def hangup_acpi(self, source, cond):
        log("acpid disconnected")
        gobject.timeout_add(10000, self.try_reconnect)
        return False
    
    """Try to reconnect -- if successful, remove timeout"""
    def try_reconnect(self):
        log("trying to reconnect to acpid")
        try: self.connect()
        except socket.error: return True
        return False
        

"""Eee PC hardware actions"""
class EeeActions(dbus.service.Object):
    def __init__(self, conn, manu, model, acpi_base, brn_path, bt_path, wlan_path, hwkey_control, obj_path="/org/eee/Eee/EeePc"):
        dbus.service.Object.__init__(self, conn, obj_path)
        self.fakekey_dev = open(glob.glob("/dev/input/by-path/platform-*-kbd")[0], "w")
        
        self.fsb_control = self.get_ovclock_control()

        if ichsmbus.smbus_init() == False:
            self.fsb_control = False
        else:
            if ichsmbus.smbus_read_block(0x69, 0) == False:
                self.fsb_control = False
        
        if self.fsb_control:
            data = ichsmbus.smbus_read_block(0x69, 0)
            # Initialize PLL to readback the maximum possible length
            data[8] = 0xff
            ichsmbus.smbus_write_block(PLL_ADDR, 0, data)
        
        # Set up EC address space
        ioport.ioperm(EC_BASE, 4, 1)
            
        # We don't need to deal with the sound/brightness keys anymore
        # when eeepc_laptop is used
        self.hwkey_control = hwkey_control

        self.acpi_base = acpi_base
        self.brn_path = brn_path
        self.bt_path = bt_path
	self.bt_path = "/sys/devices/platform/eeepc/bt"
        self.wlan_path = wlan_path

        # The wlan module and device are populated by the model definitions
        self.wlan_module = ""
        self.wlan_interface = ""

        self.model_string = model
        self.manufacturer_string = manu
        # Initialise model
        my_model = models.MODEL_MAP[self.model_string]
        self.model = my_model(self)
        self.model.wlan_path = self.wlan_path

        # Initialize state
        self.brn_state = self.get_brightness() + 0x20
        self.brn_ext_state = 0

        self.fan_control = FAN_OPTIMIZED
        self.fan_timeout = False
        
        # Fan control?
        if self.fan_control: self.start_fan_control()
        

        # If no known FSB preset is activated, apply "normal" preset
        # FIXME: really save? it may crash the machine with wrong settings...
        if self.get_fsb() == "":
            self.set_fsb("normal")

    """Generalized on/off switching"""
    @dbus.service.method("org.eee.Eee")
    def set_feature(self, what, status):
        if what in self.model.features:
            if status == True:
                what = what + "_on"
            else:
                what = what + "_off"
            try:
                m = getattr(self, what)
            except AttributeError:
                return False
            m()
            return True
        else:
            return False
            
    """Execute an action according to ACPI keycode"""
    def do_action(self, code):
        # The screen switch button runs through 0x30..0x32
        if code in (0x30, 0x31, 0x32): code = 0x30
        # On/Off hardware switches
        if code in self.model.action_map.keys():
            log("executing action %s" %self.model.action_map[code].__name__)
            self.model.action_map[code]()
        # Brightness
        elif code in range(0x20, 0x30):
            if self.brn_state < code:
                self.brn_up()
                self.brn_state = code
            elif self.brn_state > code:
                self.brn_down()
                self.brn_state = code
            # Brightness overdrive
            elif code == 0x20 and self.brn_state == 0x20:
                self.brn_down()
            elif code == 0x2f and self.brn_state == 0x2f:
                self.brn_up()
        # Programmable hotkeys
        elif code in self.model.hotkey_map:
            self.hotkey_pressed(code, self.model.hotkey_map[code])

    """ Fake a keypress """
    @dbus.service.method("org.eee.Eee")
    def fakekey(self, key):
        b = struct.pack("xxxxxxxxHHixxxxxxxxHHi", 1, key, 1, 1, key, 0)
        self.fakekey_dev.write(b)
        self.fakekey_dev.flush()
        
    """Initialize brightness from current value"""
    def get_brightness(self):
        if not os.path.exists(self.brn_path):
            log("cannot access brightness path: %s" %self.brn_path)
            exit(1)

        f = open(self.brn_path)
        l = f.readline()
        brn = int(l.strip()) + 0x20
        f.close()
        log("brightness state is %x" %brn)
        return brn
    
    """Get back a boolean value from a file in /proc"""
    def get_bool(self, path):
        f = open(path)
        l = f.readline()
	if l == "": return False
        b = bool(int(l.strip()))
        f.close()
        return b
    
    """Fan control with sigmoid function"""
    def start_fan_control(self):
        log("starting automatic fan control")
        self.fan_running = False
        # Be careful to always restore automatic adjustment when exiting!
        for s in (signal.SIGTERM, signal.SIGHUP, signal.SIGINT):
            signal.signal(s, self.fan_safe_exit)
        if self.fan_timeout: return # Already running...
        # Intialize with a low fan speed
        self.set_fan_manual(25)
        self.update_fan()
        self.fan_timeout = gobject.timeout_add(FAN_TIMEOUT, self.update_fan)
    
    def stop_fan_control(self):
        log("stopping automatic fan control")
        if self.fan_timeout: gobject.source_remove(self.fan_timeout)
        self.fan_timeout = False
        self.set_fan_auto()
        for s in (signal.SIGTERM, signal.SIGHUP, signal.SIGINT):
            signal.signal(s, signal.SIG_DFL)
    
    def fan_safe_exit(self, x, y):
        log("fan control safe exit")
        self.set_fan_auto()
        sys.exit(0)
    
    def update_fan(self):
        temp = ec_read(EC_REG_TEMP)
        speed = int(FAN_MAX_SPEED*1.0/(1+math.exp(-(temp-FAN_CRITICAL)/FAN_INTERVAL)))
        log("calculated fan speed %d" %speed)
        # Everything below 15 is considered off
        if speed < 15:
            speed = 0
            self.fan_running = False
        # Below 20 we will "kick the fan into gear"
        # by spinning it up for a second
        #  -- but only if it isn't already running
        elif speed < 20 and self.fan_running == False:
            log("kicking fan")
            self.fan_running = True
            gobject.timeout_add(1000, self.fan_kick, speed)
            speed = 30
        ec_write(EC_REG_FANPWM, speed)
        return True
    
    """Return fan to the real speed after kick-into-gear"""
    def fan_kick(self, speed):
        # if fan control has been stopped in the meantime, do nothing
        # ...but remember to unregister the timeout
        if self.fan_timeout == False:
            return False
        ec_write(EC_REG_FANPWM, speed)
        # The timeout has to be unregistered!
        return False
        
    """Emit a signal that a hotkey was pressed for the frontend"""
    @dbus.service.signal("org.eee.Eee")
    def hotkey_pressed(self, hotkey, descr):
        log("hotkey is %s" %descr)

    @dbus.service.method("org.eee.Eee")
    def mute(self):
        if self.hwkey_control:
            self.fakekey(VOLUME_MUTE)

    @dbus.service.method("org.eee.Eee")
    def vol_down(self):
        if self.hwkey_control:
            self.fakekey(VOLUME_DOWN)

    @dbus.service.method("org.eee.Eee")
    def vol_up(self):
        if self.hwkey_control:
            self.fakekey(VOLUME_UP)

    def brn_down(self):
        log("lowering brightness")
        if self.brn_state == 0x20 and self.model.extended_brightness:
            log("applying superlow brightness")
            self.brn_ext_state = -1
            self.pci_config("0000:00:02.1", 0xf4, self.model.brightness_superlow)
        elif self.brn_ext_state == 1:
            self.brn_ext_state = 0
        else:
            self.fakekey(BRIGHTNESS_DOWN)

    def brn_up(self):
        log("upping brightness")
        if self.brn_state == 0x2f and self.model.extended_brightness:
            log("applying superhigh brightness")
            self.brn_ext_state = 1
            self.pci_config("0000:00:02.1", 0xf4, self.model.brightness_superhigh)
        elif self.brn_ext_state == -1:
            self.brn_ext_state = 0
        else:
            self.fakekey(BRIGHTNESS_UP)
    
    """Modify a register in the PCI config space. Be careful here."""
    def pci_config(self, device, register, value):
        f = open("/sys/bus/pci/devices/%s/config" %device, "r")
        buf = f.read(256)
        f.close()
        bufs = [c for c in buf]
        bufs[register] = chr(value)
        buf = "".join(bufs)
        f = open("/sys/bus/pci/devices/%s/config" %device, "w")
        f.write(buf)
        f.close()

    @dbus.service.method("org.eee.Eee")
    def wifi_toggle(self):
        if self.get_bool( self.wlan_path ):
            self.model.wifi_off()
            self.wifi_changed(False)
        else:
            self.model.wifi_on()
            self.wifi_changed(True)

    # These actions are not accessible through default hardware keys    
    @dbus.service.method("org.eee.Eee")
    def bt_off(self):
        f = open(self.bt_path, "w")
        f.write("0\n")
        f.close()
        self.bluetooth_changed(False)

    @dbus.service.method("org.eee.Eee")
    def bt_on(self):
        f = open(self.bt_path, "w")
        f.write("1\n")
        f.close()
        self.bluetooth_changed(True)

    @dbus.service.method("org.eee.Eee")
    def cam_off(self):
        f = open(os.path.join(self.acpi_base, "camera"), "w")
        f.write("0\n")
        f.close()
        self.camera_changed(False)

    @dbus.service.method("org.eee.Eee")
    def cam_on(self):
        f = open(os.path.join(self.acpi_base, "camera"), "w")
        f.write("1\n")
        f.close()
        self.camera_changed(True)

    @dbus.service.method("org.eee.Eee")
    def reader_off(self):
        f = open(os.path.join(self.acpi_base, "cardr"), "w")
        f.write("0\n")
        f.close()
        self.reader_changed(False)

    @dbus.service.method("org.eee.Eee")
    def reader_on(self):
        f = open(os.path.join(self.acpi_base, "cardr"), "w")
        f.write("1\n")
        f.close()
        self.reader_changed(True)
    
    """Set FSB from userspace instead of using eee.ko"""
    @dbus.service.method("org.eee.Eee")
    def set_fsb(self, preset):
        if preset not in self.model.fsb_presets_user.keys():
            return False
        preset_data = self.model.fsb_presets_user[preset]
        pll_data = ichsmbus.smbus_read_block(PLL_ADDR, 0)
        pll_data[1] |= (1<<0) | (1<<4) | (1<<6)
        if len(pll_data) < 17:
            log("we didn't get enough pll data!")
            return False
        # M is always 24 for simplicity
        # Adjust N for the new M values
        pll_data[12] = int(24.0/(pll_data[11] & 0x3f) * pll_data[12])
        pll_data[16] = int(24.0/(pll_data[15] & 0x3f) * pll_data[16])
        pll_data[11] = 24
        pll_data[15] = 24
        curr_fsb = pll_data[12]
        curr_pci = pll_data[16]
        target_fsb, target_pci, voltage_flag = preset_data
        if curr_fsb == target_fsb and curr_pci == target_pci: return True
        log("adjustment from %d/%d to %d/%d" %
            (curr_fsb, curr_pci, target_fsb, target_pci))
        # calculate fsb steps
        fsb_dir = self.model.fsb_stepwidth
        pci_dir = self.model.fsb_stepwidth
        # direction
        if curr_fsb > target_fsb: fsb_dir = -fsb_dir
        if curr_pci > target_pci: pci_dir = -pci_dir
        fsb_steps = range(curr_fsb, target_fsb, fsb_dir)
        pci_steps = range(curr_pci, target_pci, pci_dir)
        # add target
        fsb_steps.append(int(target_fsb))
        pci_steps.append(int(target_pci))
        # normalize lists to the same length
        normalize(fsb_steps, pci_steps)
        log("FSB steps: %s" %fsb_steps)
        log("PCI steps: %s" %pci_steps)
        # apply steps
        # set voltage to high during transition
        ec_gpio_set(EC_VOLTAGE, 1)
        for fsb, pci in zip(fsb_steps, pci_steps):
            log("applying step %d/%d" %(fsb, pci))
            time.sleep(0.05)
            pll_data[12] = fsb
            ichsmbus.smbus_write_block(PLL_ADDR, 0, pll_data)
            pll_data[16] = pci
            time.sleep(0.05)
            ichsmbus.smbus_write_block(PLL_ADDR, 0, pll_data)
        # restore voltage flag
        ec_gpio_set(EC_VOLTAGE, voltage_flag)
        self.fsb_changed(preset)
        return True
        
    @dbus.service.method("org.eee.Eee")    
    def set_fan_auto(self):
        ec_write(EC_REG_FEATURE, ec_read(EC_REG_FEATURE) & ~0x02)
        self.fan_auto()

    @dbus.service.method("org.eee.Eee")    
    def set_fan_manual(self, speed):
        if speed < 0: speed = 0
        if speed > 100: speed = 100
        ec_write(EC_REG_FEATURE, ec_read(EC_REG_FEATURE) | 0x02)
        ec_write(EC_REG_FANPWM, speed)        
        self.fan_manual(speed)

    @dbus.service.method("org.eee.Eee")    
    def touchpad_off(self): 
        f = open("/sys/bus/serio/drivers/psmouse/bind_mode", "w")
        f.write("manual")
        f.close()
        f = open("/sys/bus/serio/drivers/psmouse/unbind", "w")
        f.write("serio1")
        f.close()
        self.touchpad_changed(False)

    @dbus.service.method("org.eee.Eee")    
    def touchpad_on(self):
        f = open("/sys/bus/serio/drivers/psmouse/bind_mode", "w")
        f.write("auto")
        f.close()
        self.touchpad_changed(True)
    
    @dbus.service.signal("org.eee.Eee")
    def wifi_changed(self, status): pass

    @dbus.service.signal("org.eee.Eee")
    def bluetooth_changed(self, status): pass

    @dbus.service.signal("org.eee.Eee")
    def camera_changed(self, status): pass

    @dbus.service.signal("org.eee.Eee")
    def reader_changed(self, status): pass

    @dbus.service.signal("org.eee.Eee")
    def fsb_changed(self, descr): pass
    
    @dbus.service.signal("org.eee.Eee")
    def fan_auto(self): pass

    @dbus.service.signal("org.eee.Eee")
    def fan_manual(self, speed): pass
    
    @dbus.service.signal("org.eee.Eee")
    def touchpad_changed(self, status): pass

    @dbus.service.method("org.eee.Eee")
    def get_wifi(self): return self.get_bool(self.wlan_path)

    @dbus.service.method("org.eee.Eee")
    def get_bluetooth(self): return self.get_bool(self.bt_path)

    @dbus.service.method("org.eee.Eee")
    def get_camera(self): return self.get_bool(os.path.join(self.acpi_base, "camera"))

    @dbus.service.method("org.eee.Eee")
    def get_reader(self): return self.get_bool(os.path.join(self.acpi_base, "cardr"))
    
    @dbus.service.method("org.eee.Eee")
    def get_fan(self): return bool(ec_read(EC_REG_FEATURE) & 0x02) 

    @dbus.service.method("org.eee.Eee")
    def get_fan_speed(self):
        rpm = ec_read(EC_REG_FANSPEED+1) | (ec_read(EC_REG_FANSPEED) << 8)  
        speed = ec_read(EC_REG_FANPWM)
        return speed, rpm

    @dbus.service.method("org.eee.Eee")
    def get_temperature(self):
        return ec_read(EC_REG_TEMP)   
     
    
    @dbus.service.method("org.eee.Eee")
    def get_touchpad(self): return os.path.exists("/sys/bus/serio/drivers/psmouse/serio1")

    @dbus.service.method("org.eee.Eee")
    def get_fsb(self):
        if not self.fsb_control:
            return None
        pll_data = ichsmbus.smbus_read_block(PLL_ADDR, 0)
        if len(pll_data) < 17:
            log("didn't get enough pll data!")
            return None
        p = pll_data[12]
        preset = ""
        for k, v in self.model.fsb_presets_user.iteritems():
            if v[0] == p: preset = k
        return preset

    @dbus.service.method("org.eee.Eee")
    def get_fsb_presets(self):
        presets = [(y[0], x) for x, y in self.model.fsb_presets_user.iteritems()]
        presets.sort(reverse=True)
        return [x[1] for x in presets] 
    
    @dbus.service.method("org.eee.Eee")
    def get_hotkeys(self):
        return self.model.hotkey_map
    
    @dbus.service.method("org.eee.Eee")
    def info(self):
        return (VERSION, "%s %s" %(self.manufacturer_string, self.model_string))
    
    @dbus.service.method("org.eee.Eee")
    def set_fan_control(self, state):
        if state == True and self.fan_control == False:
            self.start_fan_control()
        elif state == False and self.fan_control == True:
            self.stop_fan_control()
        self.fan_control = state
        
    @dbus.service.method("org.eee.Eee")
    def set_ovclock_control(self, state):
        self.ovclock_control = state
	controlfile = "/etc/eee-control/overclock"
	f = open(controlfile,"w")
	if state == True:
	  f.write("1")
	else:
	  f.write("0")
	f.close()

    @dbus.service.method("org.eee.Eee")
    def get_ovclock_control(self):
	controlfile = "/etc/eee-control/overclock"
	f = open(controlfile,"r")
	if f.read() == "1":
	  ret = True
	else:
	  ret = False 
	f.close()
	return ret

    @dbus.service.method("org.eee.Eee")
    def get_features(self):
        return self.model.features

