#
# Copyright (c) 2008 Grigori Goronzy <greg@geekmind.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import os

class EeePc700:
    def __init__(self, action):
        self.action = action
        self.config_section = "700"
        # FIXME: some 700s have a camera, some don't        
        self.features = "wifi", "reader", "camera", "touchpad"

        # Settings for extended brightness range.
        # superhigh is on the safe side to not kill your backlight.
        self.extended_brightness = True
        self.brightness_superlow = 0x05
        self.brightness_superhigh = 0xc0

        # This below should be the stock setting
        self.fsb_presets_user = {
            "powersave": (60, 99, 0),
            "normal": (70, 99, 0),
            "high": (85, 100, 1),
            "full": (100, 100, 1),
        }
        self.fsb_stepwidth = 2
        
        self.action_map = {
            0x10: self.action.wifi_toggle,
            0x11: self.action.wifi_toggle,
            0x13: self.action.mute,
            0x14: self.action.vol_down,
            0x15: self.action.vol_up,
        }
        
        self.hotkey_map = {
            0x30: "(Fn-F5)",
            0x12: "(Fn-F6)",
        }

        # FIXME: Check if we're using the ath5k module on Kernel 2.6.27 or
        # later. If set self.wlan_module="ath5k", self.wlan_dev="wlan0"
        self.wlan_module = "ath_pci"
        self.wlan_dev = "ath0"

    def wifi_off(self):
        # rfkill is still very flaky, at least with rt2860sta
        # it often freezes the machine often after toggling. 
        os.spawnlp(os.P_WAIT, "ifconfig", "ifconfig", self.wlan_dev, "down")
        os.spawnlp(os.P_WAIT, "modprobe", "modprobe", "-r", self.wlan_module)
        if self.wlan_path.find("rfkill") == -1: pass
        if os.path.exists(self.wlan_path):
            f = open(self.wlan_path, "w")
            f.write("0\n")
            f.close()

    def wifi_on(self):
        if os.path.exists(self.wlan_path):
            f = open(self.wlan_path, "w")
            f.write("1\n")
            f.close()
        if self.wlan_path.find("rfkill") == -1: pass
        os.spawnlp(os.P_WAIT, "modprobe", "modprobe", self.wlan_module)


class EeePc700SE(EeePc700):
    def __init__(self, action):
        EeePc700.__init__(self, action)
        self.config_section = "700SE"
        
        self.wlan_module = "r8180"
        self.wlan_dev = "wlan0"


class EeePc900(EeePc700):
    def __init__(self, action):
        EeePc700.__init__(self, action)
        self.config_section = "900"
        
        self.fsb_presets_user = {
            "full": (100, 100, 1),
            "high": (85, 100, 1),
            "normal": (70, 99, 0),
            "powersave": (60, 99, 0),
        }
        self.fsb_stepwidth = 2
        
        self.features = "wifi", "reader", "camera", "touchpad"
        self.extended_brightness = True


class EeePc900A(EeePc900):
    def __init__(self, action):
        EeePc900.__init__(self, action)
        self.config_section = "901"

        self.fsb_presets_user = {
            "normal": (48, 176, 0),
            "powersave": (38, 166, 0),
            "high": (52, 176, 1),
            "super": (56, 186, 1),
        }
        self.fsb_stepwidth = 5
        
        self.features = "wifi", "reader", "camera", "touchpad"


class EeePc901(EeePc700):
    def __init__(self, action):
        self.action = action
        self.config_section = "901"
        
        self.features = "wifi", "bt", "reader", "camera", "touchpad"
        
        # Settings for extended brightness range.
        # superhigh is on the safe side to not kill your backlight.
        self.extended_brightness = True
        self.brightness_superlow = 0x20
        self.brightness_superhigh = 0xd0
        
        # FSB/PCI-E presets for the new user-mode adjustment code
        # fsb-n, pci-n, voltage_flag
        # pci-m and pci-n are always set to 24 to
        # simplify step calculation
        self.fsb_presets_user = {
            "normal": (48, 176, 0),
            "powersave": (38, 166, 0),
            "high": (52, 176, 1),
            "super": (56, 186, 1),
        }
        self.fsb_stepwidth = 5
        
        self.action_map = {
            0x10: self.action.wifi_toggle,
            0x11: self.action.wifi_toggle,
            0x13: self.action.mute,
            0x14: self.action.vol_down,
            0x15: self.action.vol_up,
        }
        
        self.hotkey_map = {
            0x1a: "Instantkey 1",
            0x1b: "Instantkey 2",
            0x1c: "Instantkey 3",
            0x1d: "Instantkey 4",
            0x30: "(Fn-F5)",
            0x12: "(Fn-F6)",
        }

        self.wlan_module = "rt2860sta"
        self.wlan_dev = "ra0"


class EeePc1000(EeePc901):
    def __init__(self, action):
        EeePc901.__init__(self, action)
        self.config_section = "901"

        self.extended_brightness = False
        self.hotkey_map = {
            0x1a: "Instantkey 1",
            0x1b: "Instantkey 2",
            0x1c: "Instantkey 3",
            0x1d: "Instantkey 4",
            0x30: "(Fn-F8)",
            0x16: "(Fn-F7)",
            0x12: "(Fn-F9)",
        }

# Map system-product-name to the abstractions
MODEL_MAP = {
    "700": EeePc700,
    "700SE": EeePc700SE,
    "701SD": EeePc700SE,
    "701": EeePc700,
    "900": EeePc900,
    "900A": EeePc900A,
    "901": EeePc901,
    "1000": EeePc1000,
    "1000H": EeePc1000,
    "702": EeePc700,
}
